import json
import pytest
from ..history_handler import history
from awslabs.aws_api_mcp_server.core.aws.driver import translate_cli_to_ir
from awslabs.aws_api_mcp_server.core.aws.service import (
    execute_awscli_customization,
    interpret_command,
    is_operation_read_only,
    validate,
)
from awslabs.aws_api_mcp_server.core.common.command import IRCommand
from awslabs.aws_api_mcp_server.core.common.errors import AwsApiMcpError
from awslabs.aws_api_mcp_server.core.common.helpers import as_json
from awslabs.aws_api_mcp_server.core.common.models import (
    AwsCliAliasResponse,
    CommandMetadata,
    Context,
    Credentials,
    InterpretationMetadata,
    InterpretationResponse,
    InterpretedProgram,
    IRTranslation,
    ProgramInterpretationResponse,
    ValidationFailure,
)
from awslabs.aws_api_mcp_server.core.metadata.read_only_operations_list import ReadOnlyOperations
from botocore.config import Config
from tests.fixtures import (
    CLOUD9_DESCRIBE_ENVIRONMENTS,
    CLOUD9_LIST_ENVIRONMENTS,
    CLOUD9_PARAMS_CLI_MISSING_CONTEXT,
    CLOUD9_PARAMS_CLI_NON_EXISTING_OPERATION,
    CLOUD9_PARAMS_CLI_VALIDATION_FAILURES,
    CLOUD9_PARAMS_MISSING_CONTEXT_FAILURES,
    EC2_DESCRIBE_INSTANCES,
    GET_CALLER_IDENTITY_PAYLOAD,
    LAMBDA_INVOKE_PAYLOAD,
    S3_GET_OBJECT_PAYLOAD,
    SSM_LIST_NODES_PAYLOAD,
    T2_EC2_DESCRIBE_INSTANCES_FILTERED,
    TEST_CREDENTIALS,
    create_file_open_mock,
    patch_boto3,
)
from typing import Any
from unittest.mock import ANY, MagicMock, Mock, patch


@pytest.mark.parametrize(
    'cli_command,reason,service,operation',
    [
        (
            CLOUD9_PARAMS_CLI_NON_EXISTING_OPERATION,
            "The operation 'list-environments-1' for service 'cloud9' does not exist.",
            'cloud9',
            'list-environments-1',
        ),
    ],
)
def test_interpret_returns_validation_failures(cli_command, reason, service, operation):
    """Test that interpret_command returns validation failures for invalid operations."""
    response = interpret_command(
        cli_command=cli_command,
    )
    assert response.response is None
    assert response.validation_failures == [
        ValidationFailure(
            reason=reason,
            context=Context(
                service=service,
                operation=operation,
                parameters=None,
                args=None,
                region=None,
                operators=None,
            ),
        )
    ]


def test_interpret_returns_missing_context_failures():
    """Test that interpret_command returns missing context failures when required parameters are missing."""
    response = interpret_command(
        cli_command=CLOUD9_PARAMS_CLI_MISSING_CONTEXT,
    )
    assert response.response is None
    assert response.missing_context_failures == [
        ValidationFailure(
            reason="The following parameters are missing for service 'cloud9' and operation 'create-environment-ec2': '--image-id'",
            context=Context(
                service='cloud9',
                operation='create-environment-ec2',
                parameters=['--image-id'],
                args=None,
                region=None,
                operators=None,
            ),
        )
    ]


@pytest.mark.parametrize(
    'cli,output,event,service,service_full_name,operation',
    [
        (
            'aws cloud9 list-environments',
            CLOUD9_LIST_ENVIRONMENTS,
            ('ListEnvironments', {}, 'us-east-1', 60, 'https://cloud9.us-east-1.amazonaws.com'),
            'cloud9',
            'AWS Cloud9',
            'ListEnvironments',
        ),
        (
            'aws ec2 describe-instances --filters "Name=instance-state-name,Values=running"',
            EC2_DESCRIBE_INSTANCES,
            (
                'DescribeInstances',
                {
                    'Filters': [{'Name': 'instance-state-name', 'Values': ['running']}],
                },
                'us-east-1',
                60,
                'https://ec2.us-east-1.amazonaws.com',
            ),
            'ec2',
            'Amazon Elastic Compute Cloud',
            'DescribeInstances',
        ),
        (
            """aws ec2 describe-instances --query "Reservations[].Instances[?InstanceType=='t2.micro']" """,
            T2_EC2_DESCRIBE_INSTANCES_FILTERED,
            (
                'DescribeInstances',
                {},
                'us-east-1',
                60,
                'https://ec2.us-east-1.amazonaws.com',
            ),
            'ec2',
            'Amazon Elastic Compute Cloud',
            'DescribeInstances',
        ),
        (
            'aws cloud9 describe-environments --environment-ids 7d61007bd98b4d589f1504af84c168de b181ffd35fe2457c8c5ae9d75edc068a',
            CLOUD9_DESCRIBE_ENVIRONMENTS,
            (
                'DescribeEnvironments',
                {
                    'environmentIds': [
                        '7d61007bd98b4d589f1504af84c168de',  # pragma: allowlist secret
                        'b181ffd35fe2457c8c5ae9d75edc068a',  # pragma: allowlist secret
                    ]
                },
                'us-east-1',
                60,
                'https://cloud9.us-east-1.amazonaws.com',
            ),
            'cloud9',
            'AWS Cloud9',
            'DescribeEnvironments',
        ),
        (
            'aws sts get-caller-identity',
            GET_CALLER_IDENTITY_PAYLOAD,
            ('GetCallerIdentity', {}, 'us-east-1', 60, 'https://sts.us-east-1.amazonaws.com'),
            'sts',
            'AWS Security Token Service',
            'GetCallerIdentity',
        ),
        (
            'aws ssm list-nodes --sync-name Luna-Sync --filters Key=IpAddress,Values=1.0.0.1,Type=Equal',
            SSM_LIST_NODES_PAYLOAD,
            (
                'ListNodes',
                {
                    'SyncName': 'Luna-Sync',
                    'Filters': [
                        {
                            'Key': 'IpAddress',
                            'Values': ['1.0.0.1'],
                            'Type': 'Equal',
                        }
                    ],
                },
                'us-east-1',
                60,
                'https://ssm.us-east-1.amazonaws.com',
            ),
            'ssm',
            'Amazon Simple Systems Manager (SSM)',
            'ListNodes',
        ),
    ],
)
def test_interpret_returns_valid_response(
    cli, output: dict[str, Any], event, service, service_full_name, operation
):
    """Test that interpret_command returns a valid response for correct CLI commands."""
    with patch_boto3():
        with patch(
            'awslabs.aws_api_mcp_server.core.parser.parser.get_region', return_value='us-east-1'
        ):
            history.events.clear()
            response = interpret_command(cli_command=cli)
        assert response == ProgramInterpretationResponse(
            response=InterpretationResponse(json=as_json(output), error=None, status_code=200),
            failed_constraints=[],
            metadata=InterpretationMetadata(
                service=service,
                operation=operation,
                region_name='us-east-1',
                service_full_name=service_full_name,
            ),
        )
        assert event in history.events


@patch('awslabs.aws_api_mcp_server.core.parser.parser.get_region')
def test_interpret_injects_region(mock_get_region):
    """Test that interpret_command injects the correct region into the request."""
    region = 'eu-south-1'
    mock_get_region.return_value = region
    default_config = Config(region_name=region)
    with patch_boto3():
        with patch('awslabs.aws_api_mcp_server.core.parser.interpretation.Config') as patch_config:
            history.events.clear()
            patch_config.return_value = default_config
            response = interpret_command(
                cli_command='aws cloud9 describe-environments --environment-ids 7d61007bd98b4d589f1504af84c168de b181ffd35fe2457c8c5ae9d75edc068a',
            )
            assert response.metadata == InterpretationMetadata(
                service='cloud9',
                operation='DescribeEnvironments',
                region_name=region,
                service_full_name='AWS Cloud9',
            )
            event = (
                'DescribeEnvironments',
                {
                    'environmentIds': [
                        '7d61007bd98b4d589f1504af84c168de',  # pragma: allowlist secret
                        'b181ffd35fe2457c8c5ae9d75edc068a',  # pragma: allowlist secret
                    ]
                },
                'eu-south-1',
                60,
                'https://cloud9.eu-south-1.amazonaws.com',
            )
            assert event in history.events


@pytest.mark.parametrize(
    'cli, region',
    [
        (
            'aws cloudwatch list-managed-insight-rules --resource-arn arn:aws:cloudwatch:eu-west-2:123456789012:alarm:AlarmName',
            'eu-west-2',
        ),
        (
            'aws cloudwatch list-managed-insight-rules --resource-arn arn:aws:cloudwatch:eu-west-2:123456789012:alarm:AlarmName --region eu-central-1',
            'eu-central-1',
        ),
        (
            'aws cloudwatch list-managed-insight-rules --resource-arn arn:aws:cloudwatch::123456789012:alarm:AlarmName',
            'us-east-1',
        ),
    ],
)
def test_region_picked_up_from_arn(cli, region):
    """Test that region is correctly picked up from ARN in the CLI command."""
    with patch_boto3():
        with patch(
            'awslabs.aws_api_mcp_server.core.parser.parser.get_region', return_value='us-east-1'
        ):
            response = interpret_command(
                cli_command=cli,
            )
            assert response.metadata is not None
            assert response.metadata.region_name == region


def test_validate_success():
    """Test that validate returns success for a valid IR translation."""
    ir = translate_cli_to_ir('aws s3api list-buckets')
    response = validate(ir)
    response_json = json.loads(response.model_dump_json())
    assert response_json['validation_failures'] is None
    assert response_json['missing_context_failures'] is None


@pytest.mark.parametrize(
    'cli_command,validate_response',
    [
        (CLOUD9_PARAMS_CLI_NON_EXISTING_OPERATION, CLOUD9_PARAMS_CLI_VALIDATION_FAILURES),
    ],
)
def test_validate_returns_validation_failures(cli_command, validate_response):
    """Test that validate returns expected validation failures for invalid commands."""
    ir = translate_cli_to_ir(cli_command)
    response = validate(ir)
    response_json = json.loads(response.model_dump_json())
    assert response_json == validate_response


def test_validate_returns_missing_context_failures():
    """Test that validate returns missing context failures for incomplete commands."""
    ir = translate_cli_to_ir(CLOUD9_PARAMS_CLI_MISSING_CONTEXT)
    response = validate(ir)
    response_json = json.loads(response.model_dump_json())
    assert response_json == CLOUD9_PARAMS_MISSING_CONTEXT_FAILURES


@pytest.mark.parametrize(
    'cli_command,validation_failure_reason',
    [
        (
            'aws ec2 describe-instances --instance-ids abcdefgh',
            (
                "The parameter 'InstanceIds' received an invalid input: "
                'Invalid parameter value: The parameter InstanceIds does not match the ^i-[a-f0-9]{8,17}$ pattern'
            ),
        ),
        (
            'aws ec2 describe-security-groups --group-ids abcdefgh',
            (
                "The parameter 'GroupIds' received an invalid input: "
                'Invalid parameter value: The parameter GroupIds does not match the ^sg-[a-f0-9]{8,17}$ pattern'
            ),
        ),
        (
            'aws ec2 describe-instance-attribute --attribute instanceType --instance-id abcdefgh',
            (
                "The parameter 'InstanceId' received an invalid input: "
                'Invalid parameter value: The parameter InstanceId does not match the ^i-[a-f0-9]{8,17}$ pattern'
            ),
        ),
        (
            'aws ec2 describe-security-group-references --group-id abcdefgh',
            (
                "The parameter 'GroupId' received an invalid input: "
                'Invalid parameter value: The parameter GroupId does not match the ^sg-[a-f0-9]{8,17}$ pattern'
            ),
        ),
        (
            'aws ec2 revoke-security-group-ingress --group-id abcdefgh',
            (
                "The parameter 'GroupId' received an invalid input: "
                'Invalid parameter value: The parameter GroupId does not match the ^sg-[a-f0-9]{8,17}$ pattern'
            ),
        ),
    ],
)
def test_validate_returns_ec2_validation_failures(cli_command, validation_failure_reason):
    """Test that validate returns EC2 validation failures for invalid parameters."""
    ir = translate_cli_to_ir(cli_command)
    response = validate(ir)
    response_json = json.loads(response.model_dump_json())
    validation_failures = response_json['validation_failures']
    assert len(validation_failures) == 1
    assert validation_failures[0]['reason'] == validation_failure_reason


def test_is_operation_read_only_returns_true_for_read_only_operation():
    """Test is_operation_read_only returns True for a read-only operation."""
    ir = IRTranslation(
        command_metadata=CommandMetadata(
            service_sdk_name='s3',
            service_full_sdk_name='Amazon S3',
            operation_sdk_name='list-buckets',
        )
    )

    read_only_operations = ReadOnlyOperations({})
    read_only_operations['s3'] = ['list-buckets']

    result = is_operation_read_only(ir, read_only_operations)

    assert result is True


def test_is_operation_read_only_returns_false_for_non_read_only_operation():
    """Test is_operation_read_only returns False for non-read-only operation."""
    ir = IRTranslation(
        command_metadata=CommandMetadata(
            service_sdk_name='s3',
            service_full_sdk_name='Amazon S3',
            operation_sdk_name='delete-object',
        )
    )

    read_only_operations = ReadOnlyOperations({})
    read_only_operations['s3'] = ['list-buckets']

    result = is_operation_read_only(ir, read_only_operations)

    assert result is False


def test_is_operation_read_only_returns_false_for_unknown_service():
    """Test is_operation_read_only returns False for unknown service."""
    ir = IRTranslation(
        command_metadata=CommandMetadata(
            service_sdk_name='unknown-service',
            service_full_sdk_name='Unknown Service',
            operation_sdk_name='list-buckets',
        )
    )

    read_only_operations = ReadOnlyOperations({})
    read_only_operations['s3'] = ['list-buckets']

    result = is_operation_read_only(ir, read_only_operations)

    assert result is False


def test_is_operation_read_only_raises_error_for_missing_command_metadata():
    """Test is_operation_read_only raises error for missing command metadata."""
    ir = IRTranslation(command_metadata=None)
    read_only_operations = ReadOnlyOperations({})

    with pytest.raises(RuntimeError, match='failed to check if operation is allowed'):
        is_operation_read_only(ir, read_only_operations)


def test_is_operation_read_only_raises_error_for_missing_service_name():
    """Test is_operation_read_only raises error for missing service name."""
    ir = IRTranslation(
        command_metadata=CommandMetadata(
            service_sdk_name='',
            service_full_sdk_name='Amazon S3',
            operation_sdk_name='list-buckets',
        )
    )
    read_only_operations = ReadOnlyOperations({})

    with pytest.raises(RuntimeError, match='failed to check if operation is allowed'):
        is_operation_read_only(ir, read_only_operations)


def test_is_operation_read_only_raises_error_for_missing_operation_name():
    """Test is_operation_read_only raises error for missing operation name."""
    ir = IRTranslation(
        command_metadata=CommandMetadata(
            service_sdk_name='s3', service_full_sdk_name='Amazon S3', operation_sdk_name=''
        )
    )
    read_only_operations = ReadOnlyOperations({})

    with pytest.raises(RuntimeError, match='failed to check if operation is allowed'):
        is_operation_read_only(ir, read_only_operations)


@patch('awslabs.aws_api_mcp_server.core.aws.service.get_awscli_driver')
def test_execute_awscli_customization_success(mock_get_driver):
    """Test execute_awscli_customization returns AwsCliAliasResponse on successful execution."""
    mock_driver = Mock()
    mock_driver.main.return_value = None
    mock_get_driver.return_value = mock_driver

    with patch('awslabs.aws_api_mcp_server.core.aws.service.StringIO') as mock_stringio:
        mock_stdout = MagicMock()
        mock_stderr = MagicMock()
        mock_stdout.getvalue.return_value = 'bucket1\nbucket2\n'
        mock_stderr.getvalue.return_value = ''
        mock_stringio.side_effect = [mock_stdout, mock_stderr]

        cli_command = 'aws s3 ls'
        ir_command = translate_cli_to_ir(cli_command).command
        assert ir_command is not None
        result = execute_awscli_customization(cli_command, ir_command)

        assert isinstance(result, AwsCliAliasResponse)
        assert result.response == 'bucket1\nbucket2\n'
        assert result.error == ''

        mock_driver.main.assert_called_once_with(['s3', 'ls'])


@patch('awslabs.aws_api_mcp_server.core.aws.service.get_awscli_driver')
def test_execute_awscli_customization_error(mock_get_driver):
    """Test execute_awscli_customization raises AwsApiMcpError on exception."""
    mock_driver = Mock()
    mock_driver.main.side_effect = Exception('Invalid command')
    mock_get_driver.return_value = mock_driver

    with pytest.raises(AwsApiMcpError) as exc_info:
        execute_awscli_customization(
            'aws invalid command',
            IRCommand(
                command_metadata=CommandMetadata('invalid', None, 'command'),
                region='us-east-1',
                parameters={},
                is_awscli_customization=True,
            ),
        )

    assert "Error while executing 'aws invalid command': Invalid command" in str(exc_info.value)
    mock_driver.main.assert_called_once_with(['invalid', 'command'])


@patch('awslabs.aws_api_mcp_server.core.aws.service.get_awscli_driver')
@patch('awslabs.aws_api_mcp_server.core.aws.service.AWS_API_MCP_PROFILE_NAME', None)
def test_profile_not_added_when_env_var_none(mock_get_driver):
    """Test that profile is not added when AWS_API_MCP_PROFILE_NAME is None."""
    mock_driver = Mock()
    mock_get_driver.return_value = mock_driver

    cli_command = 'aws s3 ls'
    ir_command = translate_cli_to_ir(cli_command).command
    assert ir_command is not None

    execute_awscli_customization(cli_command, ir_command)

    # Verify profile was not added to args
    args = mock_driver.main.call_args[0][0]
    assert '--profile' not in args


@patch('awslabs.aws_api_mcp_server.core.aws.service.get_awscli_driver')
@patch('awslabs.aws_api_mcp_server.core.aws.service.AWS_API_MCP_PROFILE_NAME', 'test-profile')
def test_profile_added_when_env_var_set(mock_get_driver):
    """Test that profile is added when AWS_API_MCP_PROFILE_NAME is set."""
    cli_command = 'aws s3 ls'
    ir_command = translate_cli_to_ir(cli_command).command
    assert ir_command is not None
    mock_driver = Mock()
    mock_get_driver.return_value = mock_driver

    execute_awscli_customization(cli_command, ir_command)

    # Verify profile was added to args
    args = mock_driver.main.call_args[0][0]
    assert '--profile' in args
    profile_index = args.index('--profile')
    assert args[profile_index + 1] == 'test-profile'


@patch('awslabs.aws_api_mcp_server.core.aws.service.get_awscli_driver')
@patch('awslabs.aws_api_mcp_server.core.aws.service.AWS_API_MCP_PROFILE_NAME', 'test-profile')
@patch('awslabs.aws_api_mcp_server.core.parser.parser.get_region', return_value='us-east-1')
def test_profile_not_added_if_present_for_customizations(mock_get_region, mock_get_driver):
    """Test that profile is not added when one is already present."""
    cli_command = 'aws s3 ls --profile different'
    ir_command = translate_cli_to_ir(cli_command).command
    assert ir_command is not None
    mock_driver = Mock()
    mock_get_driver.return_value = mock_driver

    execute_awscli_customization(cli_command, ir_command)

    # Verify profile was added to args
    args = mock_driver.main.call_args[0][0]
    assert '--profile' in args
    profile_index = args.index('--profile')
    assert args[profile_index + 1] == 'different'


@pytest.mark.parametrize(
    'command,expected_outfile,expected_content',
    [
        (
            'aws s3api get-object --bucket test-bucket --key test-key {working_dir}/myfile.template',
            '{working_dir}/myfile.template',
            S3_GET_OBJECT_PAYLOAD['Body'].content,
        ),
        (
            'aws lambda invoke --function-name my-function {working_dir}/response.json',
            '{working_dir}/response.json',
            LAMBDA_INVOKE_PAYLOAD['Payload'].content,
        ),
    ],
)
def test_interpret_command_creates_output_file_for_streaming_operations(
    command, expected_outfile, expected_content
):
    """Test that interpret_command writes an output file for streaming operations with outfile parameter."""
    from awslabs.aws_api_mcp_server.core.common.config import WORKING_DIRECTORY

    # Replace placeholder with actual working directory
    actual_command = command.format(working_dir=WORKING_DIRECTORY)
    actual_outfile = expected_outfile.format(working_dir=WORKING_DIRECTORY)

    with patch_boto3():
        mock_open_side_effect, mock_files = create_file_open_mock(actual_outfile)

        with patch('builtins.open', side_effect=mock_open_side_effect):
            response = interpret_command(cli_command=actual_command)

            assert response.response is not None
            assert response.response.status_code == 200

            mock_file = mock_files[actual_outfile]
            mock_file.write.assert_called_with(expected_content)

            assert response.response.as_json is not None
            response_data = json.loads(response.response.as_json)

            assert 'Body' not in response_data
            assert 'Payload' not in response_data


# Tests for credentials integration changes
def test_interpret_command_with_credentials_parameter():
    """Test that interpret_command passes credentials parameter through to driver."""
    test_credentials = Credentials(**TEST_CREDENTIALS)

    with patch('awslabs.aws_api_mcp_server.core.aws.service._interpret_command') as mock_interpret:
        mock_interpret.return_value = InterpretedProgram(translation=IRTranslation())

        interpret_command('aws s3api list-buckets', credentials=test_credentials)

        mock_interpret.assert_called_once_with(
            'aws s3api list-buckets',
            max_results=None,
            credentials=test_credentials,
            default_region_override=None,
        )


def test_interpret_command_without_credentials_parameter():
    """Test that interpret_command works without credentials parameter."""
    with patch('awslabs.aws_api_mcp_server.core.aws.service._interpret_command') as mock_interpret:
        mock_interpret.return_value = InterpretedProgram(translation=IRTranslation())

        interpret_command('aws s3api list-buckets')

        mock_interpret.assert_called_once_with(
            'aws s3api list-buckets',
            max_results=None,
            credentials=None,
            default_region_override=None,
        )


@patch('awslabs.aws_api_mcp_server.core.aws.driver.interpret')
def test_interpret_command_with_region_parameter(mock_interpret):
    """Test that interpret_command forwards region to driver.interpret."""
    mock_interpret.return_value = {'ResponseMetadata': {'HTTPStatusCode': 200}}
    mock_credentials = Credentials(access_key_id='a', secret_access_key='b', session_token='c')

    interpret_command(
        'aws s3api list-buckets', default_region_override='eu-west-1', credentials=mock_credentials
    )

    mock_interpret.assert_called_once_with(
        ANY,
        access_key_id=mock_credentials.access_key_id,
        secret_access_key=mock_credentials.secret_access_key,
        session_token=mock_credentials.session_token,
        region='eu-west-1',
        client_side_filter=None,
        max_results=None,
        endpoint_url=None,
    )


@patch('awslabs.aws_api_mcp_server.core.aws.service.get_awscli_driver')
def test_execute_awscli_customization_uses_explicit_region_overrides_ir(mock_get_driver):
    """Test that execute_awscli_customization uses explicit region over IR region and default."""
    mock_driver = Mock()
    mock_get_driver.return_value = mock_driver

    ir_command = IRCommand(
        command_metadata=CommandMetadata(
            service_sdk_name='s3',
            service_full_sdk_name='Amazon S3',
            operation_sdk_name='list_objects_v2',
        ),
        parameters={},
        region='us-east-1',
        is_awscli_customization=True,
    )

    with (
        patch('awslabs.aws_api_mcp_server.core.aws.service.split_cli_command') as mock_split,
        patch('awslabs.aws_api_mcp_server.core.aws.service.operation_timer') as mock_timer,
    ):
        mock_split.return_value = ['aws', 's3', 'ls']

        # Context manager mock
        mock_cm = MagicMock()
        mock_timer.return_value = mock_cm

        with patch('sys.stdout'), patch('sys.stderr'):
            execute_awscli_customization(
                'aws s3 ls', ir_command, credentials=None, default_region_override='eu-west-2'
            )

    # Verify region precedence used in timer
    assert mock_timer.call_args[0][0] == 's3'
    assert mock_timer.call_args[0][1] == 'list_objects_v2'
    assert mock_timer.call_args[0][2] == 'us-east-1'


@patch('awslabs.aws_api_mcp_server.core.aws.service.get_awscli_driver')
def test_execute_awscli_customization_with_credentials(mock_get_driver):
    """Test that execute_awscli_customization uses provided credentials."""
    test_credentials = Credentials(**TEST_CREDENTIALS)

    mock_driver = MagicMock()
    mock_get_driver.return_value = mock_driver

    ir_command = IRCommand(
        command_metadata=CommandMetadata(
            service_sdk_name='s3',
            service_full_sdk_name='Amazon S3',
            operation_sdk_name='list_objects_v2',
        ),
        parameters={},
        region='us-east-1',
        is_awscli_customization=True,
    )

    with patch('awslabs.aws_api_mcp_server.core.aws.service.split_cli_command') as mock_split:
        mock_split.return_value = ['aws', 's3', 'ls']

        with patch(
            'awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only'
        ) as mock_is_read_only:
            mock_is_read_only.return_value = True

            with patch('sys.stdout'), patch('sys.stderr'):
                execute_awscli_customization('aws s3 ls', ir_command, credentials=test_credentials)

    mock_get_driver.assert_called_once_with(test_credentials)


@patch('awslabs.aws_api_mcp_server.core.aws.service.get_awscli_driver')
def test_execute_awscli_customization_without_credentials(mock_get_driver):
    """Test that execute_awscli_customization works without credentials."""
    mock_driver = MagicMock()
    mock_get_driver.return_value = mock_driver

    ir_command = IRCommand(
        command_metadata=CommandMetadata(
            service_sdk_name='s3',
            service_full_sdk_name='Amazon S3',
            operation_sdk_name='list_objects_v2',
        ),
        parameters={},
        region='us-east-1',
        is_awscli_customization=True,
    )

    with patch('awslabs.aws_api_mcp_server.core.aws.service.split_cli_command') as mock_split:
        mock_split.return_value = ['aws', 's3', 'ls']

        with patch(
            'awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only'
        ) as mock_is_read_only:
            mock_is_read_only.return_value = True

            with patch('sys.stdout'), patch('sys.stderr'):
                execute_awscli_customization('aws s3 ls', ir_command, credentials=None)

    mock_get_driver.assert_called_once_with(None)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import awscli.clidriver
import re
from ..common.config import get_user_agent_extra
from ..common.file_system_controls import (
    get_file_validated,
    is_streaming_blob_argument,
    validate_file_path,
)
from ..common.models import Credentials
from awscli.arguments import CLIArgument
from awscli.paramfile import URIArgumentHandler
from botocore.model import OperationModel
from collections.abc import Set
from loguru import logger
from lxml import html
from typing import Any, NamedTuple


def _deny_remote_prefix(prefix, _uri):
    raise ValueError(f'{prefix} prefix is not allowed')


RESTRICTED_URI_HANDLER = URIArgumentHandler(
    prefixes={
        'file://': (get_file_validated, {'mode': 'r'}),
        'fileb://': (get_file_validated, {'mode': 'rb'}),
        'http://': (_deny_remote_prefix, {}),
        'https://': (_deny_remote_prefix, {}),
    }
)


PaginationConfig = dict[str, int]


class ConfigResult(NamedTuple):
    """Result of configuration extraction for AWS operations."""

    parameters: dict[str, Any]
    pagination_config: PaginationConfig


filter_query = re.compile(r'^\s+([-a-z0-9_.]+|tag:<key>)\s+')


def _validate_streaming_blob_path(cli_argument: CLIArgument, value: Any, **_kwargs):
    if is_streaming_blob_argument(cli_argument) and isinstance(value, str):
        validate_file_path(value)


def get_awscli_driver(credentials: Credentials | None = None) -> awscli.clidriver.CLIDriver:
    """Create a AWS CLI driver to execute aws commands."""
    driver = awscli.clidriver.create_clidriver()
    session = driver.session
    session.register('load-cli-arg', RESTRICTED_URI_HANDLER)
    session.register('process-cli-arg.*.*', _validate_streaming_blob_path)

    # append user agent to session for aws cli customizations
    session.user_agent_extra += ' ' + get_user_agent_extra() + ' cli-customizations'
    if credentials:
        session.set_credentials(
            access_key=credentials.access_key_id,
            secret_key=credentials.secret_access_key,
            token=credentials.session_token,
        )
    return driver


class OperationFilters:
    """Represents filters for an AWS operation."""

    def __init__(self, filter_keys: Set[str], filter_set: Set[str], allows_tag_key: bool):
        """Initialize OperationFilters with filter keys, filter set, and tag key allowance."""
        self._filter_keys = frozenset(filter_keys)
        self._filter_set = frozenset(filter_set)
        self._allows_tag_key = allows_tag_key

    @property
    def filter_keys(self) -> frozenset[str]:
        """Return the set of filter keys."""
        return self._filter_keys

    def allows_filter(self, filter_name: str) -> bool:
        """Check if the given filter name is allowed."""
        if not self._filter_set:
            # Bypassing validation if filter names are not known
            return True
        return (
            filter_name in self._filter_set
            or self._allows_tag_key
            and filter_name.startswith('tag:')
        )


ALLOWED_SSM_LIST_NODES_FILTERS = {
    'AgentType',
    'AgentVersion',
    'ComputerName',
    'InstanceId',
    'InstanceStatus',
    'IpAddress',
    'ManagedStatus',
    'PlatformName',
    'PlatformType',
    'PlatformVersion',
    'ResourceType',
    'OrganizationalUnitId',
    'OrganizationalUnitPath',
    'Region',
    'AccountId',
}

# The documentation for ssm:ListDocuments doesn't list the filters
# Using the list from https://docs.aws.amazon.com/systems-manager/latest/APIReference/API_DocumentKeyValuesFilter.html
# and an undocumented key SearchKeyword
ALLOWED_SSM_LIST_DOCUMENTS_FILTERS = {
    'DocumentType',
    'Name',
    'Owner',
    'PlatformTypes',
    'SearchKeyword',
}

CUSTOM_OPERATION_FILTERS = {
    ('ssm', 'ListDocuments'): OperationFilters(
        filter_keys={'Key', 'Values'},
        filter_set=ALLOWED_SSM_LIST_DOCUMENTS_FILTERS,
        allows_tag_key=True,
    ),
    ('ssm', 'ListNodes'): OperationFilters(
        filter_keys={'Key', 'Values', 'Type'},
        filter_set=ALLOWED_SSM_LIST_NODES_FILTERS,
        allows_tag_key=True,
    ),
    ('ssm', 'ListNodesSummary'): OperationFilters(
        filter_keys={'Key', 'Values', 'Type'},
        filter_set=ALLOWED_SSM_LIST_NODES_FILTERS,
        allows_tag_key=True,
    ),
}


def get_operation_filters(operation: OperationModel) -> OperationFilters:
    """Given an operation, find all its filters."""
    filters = operation.input_shape._shape_model.get('members', {}).get('Filters')  # type: ignore[attr-defined]

    if not filters or 'documentation' not in filters:
        return OperationFilters(filter_keys=set(), filter_set=set(), allows_tag_key=False)

    if (operation.service_model.service_name, operation.name) in CUSTOM_OPERATION_FILTERS:
        return CUSTOM_OPERATION_FILTERS[
            (str(operation.service_model.service_name), str(operation.name))
        ]

    filter_keys = set()
    filters_shape = operation.service_model.shape_for(filters['shape'])
    # Single (non-list) filter validation isn't implemented right now
    if filters_shape.type_name == 'list':
        filters_shape_member = filters_shape.member
        if filters_shape_member.type_name == 'structure':
            filter_keys = set(filters_shape_member.members)  # type: ignore[attr-defined]

    # Filters are not exposed as their own field in the boto3 model, but they are
    # part of the documentation.
    filter_documentation = filters['documentation']
    filter_set = set()
    allows_tag_key = False
    for list_item in html.fromstring(filter_documentation).xpath('ul/li'):
        matched = filter_query.search(list_item.text_content())
        if matched is not None:
            filter_name = matched.group(1)
            if filter_name == 'tag:<key>':
                allows_tag_key = True
            else:
                filter_set.add(filter_name)
    if not filter_set:
        logger.warning(
            f'Empty filter set for {operation.service_model.service_name}:{operation.name}. '
            'Filter validation is likely to fail'
        )
    return OperationFilters(filter_keys, filter_set, allows_tag_key)


def extract_pagination_config(
    parameters: dict[str, Any],
    max_results: int | None = None,
) -> ConfigResult:
    """Extract pagination configuration from parameters."""
    pagination_config = parameters.pop('PaginationConfig', {})

    if max_results is None:
        return ConfigResult(parameters, pagination_config)

    max_items = pagination_config.pop('MaxItems', None)

    if max_items is not None:
        max_items = min(int(max_items), max_results)
    else:
        max_items = max_results

    pagination_config['MaxItems'] = max_items
    return ConfigResult(parameters, pagination_config)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import boto3
import importlib.metadata
import os
import tempfile
from enum import Enum
from fastmcp.server.dependencies import get_context
from loguru import logger
from pathlib import Path
from typing import Literal, cast


# Get package version for user agent
try:
    PACKAGE_VERSION = importlib.metadata.version('awslabs.aws_api_mcp_server')
except importlib.metadata.PackageNotFoundError:
    PACKAGE_VERSION = 'unknown'

TRUTHY_VALUES = frozenset(['true', 'yes', '1'])
READ_ONLY_KEY = 'READ_OPERATIONS_ONLY'
TELEMETRY_KEY = 'AWS_API_MCP_TELEMETRY'
REQUIRE_MUTATION_CONSENT_KEY = 'REQUIRE_MUTATION_CONSENT'
FILE_ACCESS_MODE_KEY = 'AWS_API_MCP_ALLOW_UNRESTRICTED_LOCAL_FILE_ACCESS'


class FileAccessMode(str, Enum):
    """File access control modes for the MCP server."""

    UNRESTRICTED = 'true'
    WORKDIR = 'workdir'
    NO_ACCESS = 'no-access'


def get_region(profile_name: str | None = None) -> str:
    """Get the region depending on configuration."""
    if AWS_REGION:
        return AWS_REGION

    fallback_region = 'us-east-1'

    if profile_name:
        return boto3.Session(profile_name=profile_name).region_name or fallback_region

    return boto3.Session().region_name or fallback_region


def get_server_directory():
    """Get platform-appropriate log directory."""
    base_location = 'aws-api-mcp'
    if os.name == 'nt' or os.uname().sysname == 'Darwin':  # Windows and macOS
        return Path(tempfile.gettempdir()) / base_location
    # Linux
    base_dir = (
        os.environ.get('XDG_RUNTIME_DIR') or os.environ.get('TMPDIR') or tempfile.gettempdir()
    )
    return Path(base_dir) / base_location


def get_env_bool(env_key: str, default: bool) -> bool:
    """Get a boolean value from an environment variable, with a default."""
    return os.getenv(env_key, str(default)).casefold() in TRUTHY_VALUES


def get_file_access_mode() -> FileAccessMode:
    """Parse the file access mode from environment variable."""
    value = os.getenv(FILE_ACCESS_MODE_KEY, 'workdir').casefold()

    # Map boolean-like values for backward compatibility
    if value in ['unrestricted', 'true', 'yes', '1']:
        return FileAccessMode.UNRESTRICTED
    elif value in ['false', 'no', '0', 'workdir']:
        return FileAccessMode.WORKDIR
    elif value == 'no-access':
        return FileAccessMode.NO_ACCESS
    else:
        # Default to workdir for unknown values
        return FileAccessMode.WORKDIR


def get_transport_from_env() -> Literal['stdio', 'streamable-http']:
    """Get a transport value from an environment variable, with a default."""
    transport = os.getenv('AWS_API_MCP_TRANSPORT', 'stdio')
    if transport not in ['stdio', 'streamable-http']:
        raise ValueError(f'Invalid transport: {transport}')

    # Enforce explicit auth configuration for streamable-http transport
    if transport == 'streamable-http':
        auth_type = os.getenv('AUTH_TYPE')
        if auth_type != 'no-auth':
            error_message = "Invalid configuration: 'streamable-http' transport requires AUTH_TYPE environment variable to be explicitly set to 'no-auth'."
            logger.error(error_message)
            raise ValueError(error_message)

    return cast(Literal['stdio', 'streamable-http'], transport)


def get_user_agent_extra() -> str:
    """Get the user agent extra string."""
    user_agent_extra = f'awslabs/mcp/AWS-API-MCP-server/{PACKAGE_VERSION}'

    try:
        ctx = get_context()
        user_agent_extra += f' via/{ctx.fastmcp.name}'

        if client_params := ctx.session.client_params:
            user_agent_extra += (
                f' MCPClient/{client_params.clientInfo.name}#{client_params.clientInfo.version}'
            )
    except RuntimeError:
        pass  # get_context throws a RuntimeError when called outside of a server request, we can safely ingore that

    if not OPT_IN_TELEMETRY:
        return user_agent_extra
    user_agent_extra += f' cfg/ro#{"1" if READ_OPERATIONS_ONLY_MODE else "0"}'
    user_agent_extra += f' cfg/consent#{"1" if REQUIRE_MUTATION_CONSENT else "0"}'
    user_agent_extra += f' cfg/scripts#{"1" if ENABLE_AGENT_SCRIPTS else "0"}'
    return user_agent_extra


FASTMCP_LOG_LEVEL = os.getenv('FASTMCP_LOG_LEVEL', 'INFO')
AWS_API_MCP_PROFILE_NAME = os.getenv('AWS_API_MCP_PROFILE_NAME')
AWS_REGION = os.getenv('AWS_REGION')
DEFAULT_REGION = get_region(AWS_API_MCP_PROFILE_NAME)
READ_OPERATIONS_ONLY_MODE = get_env_bool(READ_ONLY_KEY, False)
OPT_IN_TELEMETRY = get_env_bool(TELEMETRY_KEY, True)
WORKING_DIRECTORY = os.getenv('AWS_API_MCP_WORKING_DIR', get_server_directory() / 'workdir')
REQUIRE_MUTATION_CONSENT = get_env_bool(REQUIRE_MUTATION_CONSENT_KEY, False)
ENABLE_AGENT_SCRIPTS = get_env_bool('EXPERIMENTAL_AGENT_SCRIPTS', False)
TRANSPORT = get_transport_from_env()
HOST = os.getenv('AWS_API_MCP_HOST', '127.0.0.1')
PORT = int(os.getenv('AWS_API_MCP_PORT', 8000))
ALLOWED_HOSTS = os.getenv('AWS_API_MCP_ALLOWED_HOSTS', HOST)
ALLOWED_ORIGINS = os.getenv('AWS_API_MCP_ALLOWED_ORIGINS', HOST)
STATELESS_HTTP = get_env_bool('AWS_API_MCP_STATELESS_HTTP', False)
CUSTOM_SCRIPTS_DIR = os.getenv('AWS_API_MCP_AGENT_SCRIPTS_DIR')
FILE_ACCESS_MODE = get_file_access_mode()
ENDPOINT_SUGGEST_AWS_COMMANDS = os.getenv(
    'ENDPOINT_SUGGEST_AWS_COMMANDS', 'https://api-mcp.global.api.aws/suggest-aws-commands'
)
CONNECT_TIMEOUT_SECONDS = 10
READ_TIMEOUT_SECONDS = 60



from typing import Any


class Boto3HistoryHandler:
    """Holds events emitted during boto3 mocked calls."""

    def __init__(self):
        """Initialize the Boto3HistoryHandler with an empty events list."""
        self.events = []

    def emit(
        self,
        operation_name: str,
        payload: dict[str, Any],
        region: str,
        timeout: float,
        endpoint_url: str | None,
    ):
        """Record an emitted event with operation details."""
        self.events.append((operation_name, payload, region, timeout, endpoint_url))


history = Boto3HistoryHandler()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for file access validation in AWS CLI command parsing.

File paths appear in AWS CLI commands in the following ways:

1. Required positional outfile arguments for operations with streaming output:
   - Example: `aws s3api get-object --bucket test-bucket --key test-key /path/to/file.txt`
   - Example: `aws lambda invoke --function-name my-function /path/to/response.json`
   - Validated in: `parser.py` via `_validate_outfile()` function

2. Blob-type arguments that accept file paths with `file://` or `fileb://` prefixes:
   - Example: `aws apigatewayv2 import-api --body file:///path/to/api.yaml`
   - Example: `aws lambda invoke --function-name test --payload file:///path/to/data.json -`
   - Example: `aws rekognition detect-text --image-bytes fileb:///path/to/image.jpg`
   - Validated in: `services.py` via `RESTRICTED_URI_HANDLER` registered with 'load-cli-arg' event

3. Streaming blob-type arguments that only accept file paths without URI prefixes:
   - Example: `aws s3api put-object --bucket bucket --key file.txt --body /path/to/file.txt`
   - Example: `aws s3api upload-part --bucket bucket --key file.txt --body /path/to/part.bin --part-number 1 --upload-id x`
   - Validated in: `services.py` via `_validate_streaming_blob_path()` registered with 'process-cli-arg.*.*' event

4. Path arguments in AWS CLI customizations:
   - Example: `aws s3 cp /path/to/file.txt s3://bucket/key`
   - Example: `aws s3 sync s3://bucket/prefix /path/to/folder`
   - Example: `aws cloudformation package --template-file /home/user/cfn/template.json --s3-bucket my-bucket`
   - Example: `aws emr create-cluster --release-label emr-5.30.0 --configurations file:///tmp/config.json --instance-type m5.xlarge --instance-count 3`
   - Configured in: `file_system_controls.py` via `CUSTOM_FILE_PATH_ARGUMENTS` and `CUSTOM_BLOB_ARGUMENTS`
   - Validated in: `parser.py` via `_validate_customization_file_paths()` function

File Access Control Modes:

The FILE_ACCESS_MODE configuration controls file system access through three modes:

- FileAccessMode.WORKDIR (default): Only file paths within WORKING_DIRECTORY are allowed.
  This provides a sandboxed environment for file operations.

- FileAccessMode.UNRESTRICTED: Allows file paths anywhere on the local file system,
  bypassing the working directory restriction.

- FileAccessMode.NO_ACCESS: Completely denies the use of local file paths in commands.
  S3 URIs (s3://...) and stdout redirect (-) remain allowed.
"""

import pytest
from awslabs.aws_api_mcp_server.core.common.config import WORKING_DIRECTORY, FileAccessMode
from awslabs.aws_api_mcp_server.core.common.errors import (
    FileParameterError,
    FilePathValidationError,
    LocalFileAccessDisabledError,
    OperationNotAllowedError,
)
from awslabs.aws_api_mcp_server.core.parser.parser import parse
from pathlib import Path
from unittest.mock import patch


@patch(
    'awslabs.aws_api_mcp_server.core.parser.parser.FILE_ACCESS_MODE',
    FileAccessMode.WORKDIR,
)
@patch(
    'awslabs.aws_api_mcp_server.core.common.file_system_controls.FILE_ACCESS_MODE',
    FileAccessMode.WORKDIR,
)
class TestDefaultFileAccessBehavior:
    """Tests for default file access validation behavior."""

    @patch(
        'awslabs.aws_api_mcp_server.core.common.file_system_controls.WORKING_DIRECTORY',
        Path.home(),
    )
    def test_valid_expand_user_home_directory(self):
        """Test that tilde or user home directory is invalid path."""
        result = parse(cli_command='aws s3 cp s3://my-bucket/my_file ~/temp/test.txt')
        assert result.command_metadata.service_sdk_name == 's3'
        assert result.command_metadata.operation_sdk_name == 'cp'
        assert len(result.parameters['--paths']) == 2
        assert result.parameters['--paths'][1] == str(Path.home() / 'temp' / 'test.txt')

    @patch(
        'awslabs.aws_api_mcp_server.core.common.file_system_controls.WORKING_DIRECTORY',
        Path.home(),
    )
    def test_valid_expand_user_home_directory_in_outfile_arg(self):
        """Test that tilde or user home directory is invalid path."""
        result = parse(
            cli_command='aws s3api get-object --bucket bucket --key file.txt ~/file.txt'
        )
        assert result.command_metadata.service_sdk_name == 's3'
        assert result.command_metadata.operation_sdk_name == 'GetObject'
        assert result.output_file is not None
        assert result.output_file.path == str(Path.home() / 'file.txt')

    @patch(
        'awslabs.aws_api_mcp_server.core.common.file_system_controls.WORKING_DIRECTORY',
        Path.home(),
    )
    def test_invalid_expand_user_home_directory(self):
        """Test that tilde is not expanded."""
        expected_message = (
            "Invalid file parameter '~user_that_does_not_exist/temp/test.txt' for service 's3' and operation 'cp': "
            'should be an absolute path.'
        )
        with pytest.raises(FileParameterError) as exc_info:
            parse(cli_command='aws s3 cp s3://my_file ~user_that_does_not_exist/temp/test.txt')

        assert str(exc_info.value) == expected_message

    @pytest.mark.parametrize(
        'command,expected_service,expected_operation,expected_file_path',
        [
            (
                'aws s3api get-object --bucket test-bucket --key test-key relative/path/file.txt',
                's3',
                'GetObject',
                'relative/path/file.txt',
            ),
            (
                'aws lambda invoke --function-name my-function response.json',
                'lambda',
                'Invoke',
                'response.json',
            ),
        ],
    )
    def test_validate_output_file_raises_error_for_relative_paths(
        self, command, expected_service, expected_operation, expected_file_path
    ):
        """Test that _validate_output_file raises FileParameterError for streaming operations with relative paths."""
        expected_message = f"Invalid file parameter '{expected_file_path}' for service '{expected_service}' and operation '{expected_operation}': should be an absolute path"
        with pytest.raises(FileParameterError, match=expected_message):
            parse(command)

    @pytest.mark.parametrize(
        'command',
        [
            'aws lambda invoke --function-name MyFunction --payload file://relative/path/payload.json -',
            'aws rekognition detect-text --image-bytes fileb://relative/path/test.jpg',
            'aws s3api put-object --bucket bucket --key file.txt --body pyproject.toml',
        ],
    )
    def test_parse_raises_error_for_relative_paths_in_blob_args(self, command):
        """Test that _validate_output_file raises CommandValidationError for blob args with relative paths."""
        with pytest.raises(
            FilePathValidationError,
            match='is outside the allowed working directory',
        ):
            parse(command)

    @pytest.mark.parametrize(
        'command',
        [
            # S3 custom file path arguments outside WORKING_DIRECTORY
            'aws s3 cp /tmp/file.txt s3://bucket/key',
            'aws s3 cp s3://bucket/key /tmp/file.txt',
            'aws s3 sync /tmp/folder s3://bucket/prefix',
            'aws s3 mv /tmp/file.txt s3://bucket/key',
            # CloudFormation custom file path arguments outside WORKING_DIRECTORY
            'aws cloudformation package --template-file /tmp/template.yaml --s3-bucket my-bucket',
            f'aws cloudformation package --template-file {WORKING_DIRECTORY}/template.yaml --s3-bucket my-bucket --output-template-file /tmp/packaged.yaml',
            'aws cloudformation deploy --template-file /tmp/template.json --stack-name my-stack',
            # CloudFront custom file path arguments outside WORKING_DIRECTORY
            'aws cloudfront sign --url https://example.com --private-key /tmp/private-key.pem --key-pair-id APKAEXAMPLE --date-less-than 2025-12-31',
            # ECS custom file path arguments outside WORKING_DIRECTORY
            'aws ecs deploy --cluster my-cluster --service my-service --task-definition /tmp/task-def.json --codedeploy-appspec /tmp/appspec.yaml',
            # EKS custom file path arguments outside WORKING_DIRECTORY
            'aws eks update-kubeconfig --name my-cluster --kubeconfig /tmp/kubeconfig',
            # GameLift custom file path arguments outside WORKING_DIRECTORY
            'aws gamelift upload-build --name my-build --build-version 1.0 --build-root /tmp/build --operating-system AMAZON_LINUX_2',
            'aws gamelift get-game-session-log --game-session-id gsess-123 --save-as /tmp/log.txt',
            # Deploy custom file path arguments outside WORKING_DIRECTORY
            'aws deploy push --application-name my-app --s3-location s3://my-bucket/app.zip --source /tmp/app',
            # EMR custom blob arguments outside WORKING_DIRECTORY
            'aws emr create-cluster --name my-cluster --release-label emr-5.30.0 --configurations file:///tmp/config.json --instance-type m5.xlarge --instance-count 3',
            'aws emr create-cluster --name my-cluster --release-label emr-5.30.0 --configurations InstanceCount=3,InstanceGroupType=MASTER --instance-fleets file:///tmp/fleets.json',
            'aws emr add-steps --cluster-id j-123456 --steps file:///tmp/steps.json',
        ],
    )
    def test_custom_file_path_arguments_outside_working_directory_rejected(self, command):
        """Test that custom file path arguments outside WORKING_DIRECTORY are rejected."""
        with pytest.raises(
            FileParameterError,
            match='is outside the allowed working directory',
        ):
            parse(command)


@patch(
    'awslabs.aws_api_mcp_server.core.parser.parser.FILE_ACCESS_MODE',
    FileAccessMode.NO_ACCESS,
)
@patch(
    'awslabs.aws_api_mcp_server.core.common.file_system_controls.FILE_ACCESS_MODE',
    FileAccessMode.NO_ACCESS,
)
class TestDisabledLocalFileAccess:
    """Tests for when local file access is disabled via FILE_ACCESS_MODE.NO_ACCESS."""

    @pytest.mark.parametrize(
        'command',
        [
            f'aws s3 cp {WORKING_DIRECTORY}/file.txt s3://bucket/key',
            f'aws s3 cp s3://bucket/key {WORKING_DIRECTORY}/file.txt',
            f'aws s3 sync {WORKING_DIRECTORY}/folder s3://bucket/prefix',
            f'aws s3 mv {WORKING_DIRECTORY}/file.txt s3://bucket/key',
            f'aws s3api get-object --bucket bucket --key file.txt {WORKING_DIRECTORY}/file.txt',
            f'aws lambda invoke --function-name MyFunction {WORKING_DIRECTORY}/result.json',
            f'aws emr add-steps --cluster-id j-ABCD1234EFGH --steps file://{WORKING_DIRECTORY}/steps.json',
        ],
    )
    def test_rejects_commands_with_local_paths(self, command):
        """Test that commands with local paths are rejected when FILE_ACCESS_MODE is NO_ACCESS."""
        with pytest.raises(
            FileParameterError,
            match='local file access is disabled',
        ):
            parse(command)

    @pytest.mark.parametrize(
        'command',
        [
            f'aws lambda invoke --function-name MyFunction --payload file://{WORKING_DIRECTORY}/payload.json -',
            f'aws lambda update-function-code --function-name MyFunction --zip-file fileb://{WORKING_DIRECTORY}/test.jpg',
            f'aws apigatewayv2 import-api --body file://{WORKING_DIRECTORY}/api.yaml',
            f'aws rekognition detect-text --image-bytes fileb://{WORKING_DIRECTORY}/test.jpg',
        ],
    )
    def test_rejects_file_blob_arguments(self, command):
        """Test that commands with file:// and fileb:// arguments are rejected when FILE_ACCESS_MODE is NO_ACCESS."""
        with pytest.raises(LocalFileAccessDisabledError):
            parse(command)

    @pytest.mark.parametrize(
        'command',
        [
            f'aws s3api put-object --bucket bucket --key file.txt --body {WORKING_DIRECTORY}/test-file.txt',
            f'aws s3api upload-part --bucket bucket --key file.txt --body {WORKING_DIRECTORY}/part.bin --part-number 1 --upload-id x',
            f'aws lambda invoke-async --function-name MyFunction --invoke-args {WORKING_DIRECTORY}/args.json',
        ],
    )
    def test_rejects_streaming_blob_arguments(self, command):
        """Test that commands with streaming blob arguments are rejected when FILE_ACCESS_MODE is NO_ACCESS.

        Streaming blob arguments accept only file paths.
        """
        with pytest.raises(LocalFileAccessDisabledError):
            parse(command)

    @pytest.mark.parametrize(
        'command,expected_service,expected_operation',
        [
            ('aws s3 cp s3://source-bucket/key s3://dest-bucket/key', 's3', 'cp'),
            ('aws s3 sync s3://source-bucket/prefix s3://dest-bucket/prefix', 's3', 'sync'),
            ('aws s3 mv s3://source-bucket/key s3://dest-bucket/key', 's3', 'mv'),
            ('aws s3api get-object --bucket bucket --key file.txt -', 's3', 'get-object'),
            (
                'aws lambda invoke --function-name MyFunction --payload \'{"key": "value"}\' -',
                'lambda',
                'invoke',
            ),
        ],
    )
    def test_allows_s3_uris_and_stdout_redirect(
        self, command, expected_service, expected_operation
    ):
        """Test that S3 URIs, stdout redirect, and inline values are allowed when FILE_ACCESS_MODE is NO_ACCESS."""
        result = parse(command)
        assert result.command_metadata.service_sdk_name == expected_service
        assert result.operation_cli_name == expected_operation

    @pytest.mark.parametrize(
        'command,expected_service,expected_operation',
        [
            # S3 operations that don't require local files
            ('aws s3 ls', 's3', 'ls'),
            ('aws s3 ls s3://bucket', 's3', 'ls'),
            ('aws s3 website s3://bucket --index-document index.html', 's3', 'website'),
            ('aws s3 rm s3://bucket/key', 's3', 'rm'),
            ('aws s3 mb s3://new-bucket', 's3', 'mb'),
            ('aws s3 rb s3://bucket', 's3', 'rb'),
            ('aws s3 presign s3://bucket/key', 's3', 'presign'),
            ('aws s3 cp s3://source/key s3://dest/key', 's3', 'cp'),
            ('aws s3 mv s3://source/key s3://dest/key', 's3', 'mv'),
            ('aws s3 sync s3://source s3://dest', 's3', 'sync'),
            # CloudTrail operations
            (
                'aws cloudtrail validate-logs --trail-arn arn:aws:cloudtrail:us-east-1:123456789012:trail/my-trail --start-time 2023-01-01T00:00:00Z',
                'cloudtrail',
                'validate-logs',
            ),
            # DataPipeline operations
            ('aws datapipeline list-runs --pipeline-id df-123456789', 'datapipeline', 'list-runs'),
            ('aws datapipeline create-default-roles', 'datapipeline', 'create-default-roles'),
            # DLM operations
            ('aws dlm create-default-role', 'dlm', 'create-default-role'),
            # ECR operations
            ('aws ecr get-login --region us-east-1 --no-include-email', 'ecr', 'get-login'),
            ('aws ecr get-login-password', 'ecr', 'get-login-password'),
            # ECR Public operations
            ('aws ecr-public get-login-password', 'ecr-public', 'get-login-password'),
            # EKS operations
            ('aws eks get-token --cluster-name my-cluster', 'eks', 'get-token'),
            # EMR operations
            (
                'aws emr add-instance-groups --cluster-id j-123456 --instance-groups InstanceGroupType=TASK,InstanceType=m5.xlarge,InstanceCount=2',
                'emr',
                'add-instance-groups',
            ),
            (
                'aws emr create-cluster --name my-cluster --release-label emr-5.30.0 --applications Name=Spark --ec2-attributes KeyName=my-key --instance-type m5.xlarge --instance-count 3 --use-default-roles',
                'emr',
                'create-cluster',
            ),
            ('aws emr describe-cluster --cluster-id j-123456', 'emr', 'describe-cluster'),
            (
                'aws emr modify-cluster-attributes --cluster-id j-123456 --visible-to-all-users',
                'emr',
                'modify-cluster-attributes',
            ),
            ('aws emr create-default-roles', 'emr', 'create-default-roles'),
            (
                'aws emr add-steps --cluster-id j-XXXXXXXX --steps Type=CUSTOM_JAR,Name=CustomJAR,ActionOnFailure=CONTINUE,Jar=s3://amzn-s3-demo-bucket/mytest.jar,Args=arg1,arg2,arg3',
                'emr',
                'add-steps',
            ),
            # EMR Containers operations
            (
                'aws emr-containers update-role-trust-policy --cluster-name my-cluster --namespace default --role-name my-role',
                'emr-containers',
                'update-role-trust-policy',
            ),
            # RDS operations
            (
                'aws rds generate-db-auth-token --hostname mydb.example.com --port 3306 --username admin',
                'rds',
                'generate-db-auth-token',
            ),
            # Deploy operations
            ('aws deploy deregister --instance-name my-instance', 'deploy', 'deregister'),
            # ConfigService operations
            ('aws configservice get-status', 'configservice', 'get-status'),
        ],
    )
    def test_allowed_custom_operations_work_with_disabled_file_access(
        self, command, expected_service, expected_operation
    ):
        """Test that allowed custom operations work when FILE_ACCESS_MODE is NO_ACCESS."""
        result = parse(command)
        assert result.command_metadata.service_sdk_name == expected_service
        assert result.operation_cli_name == expected_operation
        assert result.is_awscli_customization is True

    @pytest.mark.parametrize(
        'command,expected_service,expected_operation',
        [
            # CloudFormation operations that require local files
            (
                f'aws cloudformation package --template-file {WORKING_DIRECTORY}/template.yaml --s3-bucket my-bucket',
                'cloudformation',
                'package',
            ),
            (
                f'aws cloudformation deploy --template-file {WORKING_DIRECTORY}/template.yaml --stack-name my-stack',
                'cloudformation',
                'deploy',
            ),
            # CloudFront operations that require local files
            (
                f'aws cloudfront sign --url https://example.com --private-key {WORKING_DIRECTORY}/private-key.pem --key-pair-id APKAEXAMPLE',
                'cloudfront',
                'sign',
            ),
            # CodeArtifact operations that require local files
            (
                'aws codeartifact login --tool pip --domain my-domain --repository my-repo',
                'codeartifact',
                'login',
            ),
            # ECS operations that require local files
            (
                f'aws ecs deploy --cluster my-cluster --service my-service --task-definition {WORKING_DIRECTORY}/task-def.json',
                'ecs',
                'deploy',
            ),
            # EKS operations that require local files
            (
                'aws eks update-kubeconfig --name my-cluster',
                'eks',
                'update-kubeconfig',
            ),
            # GameLift operations that require local files
            (
                f'aws gamelift upload-build --name my-build --build-version 1.0 --build-root {WORKING_DIRECTORY}/build --operating-system AMAZON_LINUX_2',
                'gamelift',
                'upload-build',
            ),
            (
                f'aws gamelift get-game-session-log --game-session-id gsess-123 --save-as {WORKING_DIRECTORY}/log.txt',
                'gamelift',
                'get-game-session-log',
            ),
            # ServiceCatalog operations that require local files
            (
                f'aws servicecatalog generate --product-id prod-123 --provisioning-artifact-id pa-123 --output-file {WORKING_DIRECTORY}/output.json',
                'servicecatalog',
                'generate',
            ),
            # Deploy operations that require local files
            (
                f'aws deploy push --application-name my-app --s3-location s3://my-bucket/app.zip --source {WORKING_DIRECTORY}/app',
                'deploy',
                'push',
            ),
            (
                'aws deploy register --application-name my-app --s3-location s3://my-bucket/app.zip --description "My app"',
                'deploy',
                'register',
            ),
        ],
    )
    def test_file_based_custom_operations_rejected_with_disabled_file_access(
        self, command, expected_service, expected_operation
    ):
        """Test that file-based custom operations are rejected when FILE_ACCESS_MODE is NO_ACCESS."""
        with pytest.raises(OperationNotAllowedError) as exc_info:
            parse(command)

        error_message = str(exc_info.value)
        assert expected_service in error_message.lower()
        assert expected_operation in error_message.lower()


@patch(
    'awslabs.aws_api_mcp_server.core.common.file_system_controls.FILE_ACCESS_MODE',
    FileAccessMode.UNRESTRICTED,
)
class TestUnrestrictedLocalFileAccess:
    """Tests for when unrestricted local file access is enabled."""

    def test_allows_local_files_in_unrestricted_mode(self):
        """Test that FILE_ACCESS_MODE.UNRESTRICTED allows local files anywhere on the filesystem."""
        # With UNRESTRICTED mode, local files should be allowed
        command = 'aws s3 cp /tmp/file.txt s3://bucket/key'
        # Should not raise any exception about file access
        result = parse(command)
        assert result.command_metadata.service_sdk_name == 's3'



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
from ..aws.services import get_awscli_driver
from ..common.config import AWS_API_MCP_PROFILE_NAME, DEFAULT_REGION
from ..common.errors import AwsApiMcpError, Failure
from ..common.help_command import generate_help_document
from ..common.models import (
    AwsCliAliasResponse,
    Consent,
    Credentials,
    InterpretationMetadata,
    InterpretationResponse,
    InterpretedProgram,
    IRTranslation,
    ProgramInterpretationResponse,
    ProgramValidationResponse,
)
from ..common.models import Context as ContextAPIModel
from ..common.models import ValidationFailure as FailureAPIModel
from ..metadata.read_only_operations_list import (
    ReadOnlyOperations,
)
from ..parser.lexer import split_cli_command
from ..security.policy import PolicyDecision, SecurityPolicy
from .driver import interpret_command as _interpret_command
from awslabs.aws_api_mcp_server.core.common.command import IRCommand
from awslabs.aws_api_mcp_server.core.common.helpers import as_json, operation_timer
from fastmcp import Context
from fastmcp.server.elicitation import AcceptedElicitation
from io import StringIO
from loguru import logger
from mcp.shared.exceptions import McpError
from mcp.types import METHOD_NOT_FOUND
from typing import Any


async def request_consent(cli_command: str, ctx: Context):
    """Request consent of the user using elicitation."""
    try:
        elicitation_result = await ctx.elicit(
            message=f"The CLI command '{cli_command}' requires explicit consent. Do you approve the execution of this command?",
            response_type=Consent,
        )

        if (
            not isinstance(elicitation_result, AcceptedElicitation)
            or not elicitation_result.data.answer
        ):
            error_message = 'User rejected the execution of the command.'
            await ctx.error(error_message)
            raise AwsApiMcpError(error_message)
    except McpError as e:
        if e.error.code == METHOD_NOT_FOUND:
            error_message = 'Client does not support elicitation. Use a different client or update the server configuration.'
            logger.error(error_message)
            raise AwsApiMcpError(error_message)

        raise e


def is_operation_read_only(ir: IRTranslation, read_only_operations: ReadOnlyOperations):
    """Check if the operation in the IR is read-only."""
    if (
        not ir.command_metadata
        or not getattr(ir.command_metadata, 'service_sdk_name', None)
        or not getattr(ir.command_metadata, 'operation_sdk_name', None)
    ):
        raise RuntimeError(
            "failed to check if operation is allowed: translated command doesn't include service and operation name"
        )

    service_name = ir.command_metadata.service_sdk_name
    operation_name = ir.command_metadata.operation_sdk_name
    return read_only_operations.has(service=service_name, operation=operation_name)


def check_security_policy(
    ir: IRTranslation, read_only_operations: ReadOnlyOperations, ctx: Context
) -> PolicyDecision:
    """Check security policy for the given command and return decision."""

    def is_read_only_func(service: str, operation: str) -> bool:
        return read_only_operations.has(service=service, operation=operation)

    policy = SecurityPolicy(ctx)

    # First check if this matches a customization
    customization_decision = policy.check_customization(ir, is_read_only_func)
    if customization_decision is not None:
        return customization_decision

    # If no customization matches, check individual operation
    if (
        not ir.command_metadata
        or not getattr(ir.command_metadata, 'service_sdk_name', None)
        or not getattr(ir.command_metadata, 'operation_sdk_name', None)
    ):
        return PolicyDecision.ELICIT if policy.supports_elicitation else PolicyDecision.DENY

    service_name = ir.command_metadata.service_sdk_name
    operation_name = ir.command_metadata.operation_sdk_name
    is_read_only = is_operation_read_only(ir, read_only_operations)

    return policy.determine_policy_effect(service_name, operation_name, is_read_only)


def validate(ir: IRTranslation) -> ProgramValidationResponse:
    """Translate the given CLI command and return a validation response."""
    return ProgramValidationResponse(
        missing_context_failures=_to_missing_context_failures(ir.missing_context_failures),
        validation_failures=_to_validation_failures(ir.validation_or_translation_failures),
    )


async def get_help_document(
    cli_command: str,
    ctx: Context,
) -> ProgramInterpretationResponse:
    """Get help command response."""
    args = split_cli_command(cli_command)[1:]
    service_name = args[0]
    operation_name = args[1]
    help_document = generate_help_document(service_name, operation_name)
    if help_document is None:
        error_message = 'Failed to generate help document'
        await ctx.error(error_message)
        raise AwsApiMcpError(error_message)
    return ProgramInterpretationResponse(
        response=InterpretationResponse(json=as_json(help_document), status_code=200, error=None)
    )


def execute_awscli_customization(
    cli_command: str,
    ir_command: IRCommand,
    credentials: Credentials | None = None,
    default_region_override: str | None = None,
) -> AwsCliAliasResponse:
    """Execute the given AWS CLI command."""
    args = split_cli_command(cli_command)[1:]

    # Identify if a profile was passed in already and insert the defined one otherwise
    if AWS_API_MCP_PROFILE_NAME and not any(elem == '--profile' for elem in args):
        args.extend(['--profile', AWS_API_MCP_PROFILE_NAME])

    try:
        stdout_capture = StringIO()
        stderr_capture = StringIO()

        with (
            contextlib.redirect_stdout(stdout_capture),
            contextlib.redirect_stderr(stderr_capture),
        ):
            with operation_timer(
                ir_command.service_name,
                ir_command.operation_name,
                ir_command.region or default_region_override or DEFAULT_REGION,
            ):
                driver = get_awscli_driver(credentials)
                driver.main(args)

        stdout_output = stdout_capture.getvalue()
        stderr_output = stderr_capture.getvalue()

        return AwsCliAliasResponse(response=stdout_output, error=stderr_output)
    except Exception as e:
        raise AwsApiMcpError(f"Error while executing '{cli_command}': {e}")


def interpret_command(
    cli_command: str,
    max_results: int | None = None,
    credentials: Credentials | None = None,
    default_region_override: str | None = None,
) -> ProgramInterpretationResponse:
    """Interpret the given CLI command and return an interpretation response."""
    interpreted_program = _interpret_command(
        cli_command,
        max_results=max_results,
        credentials=credentials,
        default_region_override=default_region_override,
    )

    validation_failures = (
        []
        if not interpreted_program.translation.validation_or_translation_failures
        else interpreted_program.translation.validation_or_translation_failures
    )
    missing_context_failures = (
        []
        if not interpreted_program.translation.missing_context_failures
        else interpreted_program.translation.missing_context_failures
    )
    failed_constraints = interpreted_program.failed_constraints or []

    if (
        not validation_failures
        and not missing_context_failures
        and not interpreted_program.failed_constraints
    ):
        response = InterpretationResponse(
            json=interpreted_program.response,
            error=interpreted_program.service_error,
            status_code=interpreted_program.status_code,
            error_code=interpreted_program.error_code,
            pagination_token=interpreted_program.pagination_token,
        )
    else:
        response = None

    return ProgramInterpretationResponse(
        response=response,
        metadata=_ir_metadata(interpreted_program),
        validation_failures=_to_validation_failures(validation_failures),
        missing_context_failures=_to_missing_context_failures(missing_context_failures),
        failed_constraints=failed_constraints,
    )


def _ir_metadata(program: InterpretedProgram | None) -> InterpretationMetadata | None:
    if program and program.translation and program.translation.command:
        command = program.translation.command
        return InterpretationMetadata(
            service=command.service_name,
            service_full_name=command.service_full_name,
            operation=command.operation_name,
            region_name=program.region_name,
        )
    return None


def _to_missing_context_failures(
    failures: list[Failure] | None,
) -> list[FailureAPIModel] | None:
    if not failures:
        return None

    return [
        FailureAPIModel(reason=failure.reason, context=_to_context(failure.context))
        for failure in failures
    ]


def _to_validation_failures(failures: list[Failure] | None) -> list[FailureAPIModel] | None:
    if not failures:
        return None

    return [
        FailureAPIModel(reason=failure.reason, context=_to_context(failure.context))
        for failure in failures
    ]


def _to_context(context: dict[str, Any] | None) -> ContextAPIModel | None:
    if not context:
        return None

    return ContextAPIModel(
        service=context.get('service'),
        operation=context.get('operation'),
        operators=context.get('operators'),
        region=context.get('region'),
        args=context.get('args'),
        parameters=context.get('parameters'),
    )



import pytest
from awslabs.aws_api_mcp_server.core.aws.services import (
    extract_pagination_config,
    get_awscli_driver,
)
from awslabs.aws_api_mcp_server.core.common.models import Credentials
from tests.fixtures import TEST_CREDENTIALS
from unittest.mock import MagicMock, patch


MAX_RESULTS = 6


@pytest.mark.parametrize(
    'max_result_config, max_result_param, expected_max_result',
    [
        (None, None, None),  # max result is not defined
        (10, None, 10),  # max result is defined in config but not as param
        (None, 6, 6),  # max result is defined as parameter and not in config
        (6, 10, 6),  # max result is defined in both places. In this case we take config param
    ],
)
def test_max_results(max_result_config, max_result_param, expected_max_result):
    """Test that max results are set correctly based on config and parameters."""
    parameters = {
        'PaginationConfig': {'MaxItems': max_result_config},
        'Foo': 'Bar',
    }
    updated_parameters, pagination_config = extract_pagination_config(parameters, max_result_param)
    max_results = pagination_config.get('MaxItems')
    assert max_results == expected_max_result
    assert updated_parameters.get('PaginationConfig') is None


@patch('os.environ.get')
@patch('botocore.httpsession.URLLib3Session.send')
def test_session_user_agent_in_boto_request(mock_send, mock_env):
    """Test that boto requests include the MCP user agent."""
    mock_env.side_effect = lambda key, default=None: {
        'AWS_ACCESS_KEY_ID': 'test',  # pragma: allowlist secret
        'AWS_SECRET_ACCESS_KEY': 'test',  # pragma: allowlist secret
    }.get(key, default)

    mock_response = type(
        'MockResponse',
        (),
        {
            'status_code': 200,
            'headers': {},
            'content': b'<GetCallerIdentityResponse><GetCallerIdentityResult></GetCallerIdentityResult></GetCallerIdentityResponse>',
        },
    )()
    mock_send.return_value = mock_response

    session = get_awscli_driver().session
    client = session.create_client('sts', region_name='us-east-1')
    client.get_caller_identity()

    user_agent = mock_send.call_args[0][0].headers.get('User-Agent', b'').decode()
    assert 'awslabs/mcp/AWS-API-MCP-server' in user_agent
    assert 'cli-customizations' in user_agent


# Tests for get_awscli_driver function
@patch('awscli.clidriver.create_clidriver')
def test_get_awscli_driver_with_credentials(mock_create_driver):
    """Test get_awscli_driver with credentials sets them on the session."""
    test_credentials = Credentials(**TEST_CREDENTIALS)

    mock_driver = MagicMock()
    mock_session = MagicMock()
    mock_driver.session = mock_session
    mock_create_driver.return_value = mock_driver

    result = get_awscli_driver(test_credentials)

    assert result == mock_driver
    mock_session.set_credentials.assert_called_once_with(
        access_key=test_credentials.access_key_id,
        secret_key=test_credentials.secret_access_key,
        token=test_credentials.session_token,
    )


@patch('awscli.clidriver.create_clidriver')
def test_get_awscli_driver_without_credentials(mock_create_driver):
    """Test get_awscli_driver without credentials does not set credentials."""
    mock_driver = MagicMock()
    mock_session = MagicMock()
    mock_driver.session = mock_session
    mock_create_driver.return_value = mock_driver

    result = get_awscli_driver(None)

    assert result == mock_driver
    mock_session.set_credentials.assert_not_called()


@patch('awscli.clidriver.create_clidriver')
def test_get_awscli_driver_user_agent_configuration(mock_create_driver):
    """Test get_awscli_driver configures user agent correctly."""
    mock_driver = MagicMock()
    mock_session = MagicMock()
    mock_session.user_agent_extra = ''
    mock_driver.session = mock_session
    mock_create_driver.return_value = mock_driver

    result = get_awscli_driver(None)

    assert result == mock_driver
    assert 'awslabs/mcp/AWS-API-MCP-server' in mock_session.user_agent_extra
    assert 'cli-customizations' in mock_session.user_agent_extra



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from ..parser.parser import driver
from awscli.bcdoc.restdoc import ReSTDocument
from awscli.clidriver import ServiceCommand
from awscli.customizations.commands import BasicCommand
from loguru import logger
from typing import Any


IGNORED_ARGUMENTS = frozenset({'cli-input-json', 'generate-cli-skeleton'})


def _clean_text(text: str) -> str:
    text = re.sub(r'\s+', ' ', text)  # Normalize whitespace
    return text.strip()


def _clean_description(description: str) -> str:
    """This removes the section title added by the help event handlers."""
    description = re.sub(r'=+\s*Description\s*=+\s', '', description)
    return _clean_text(description)


def generate_help_document(service_name: str, operation_name: str) -> dict[str, Any] | None:
    """Generate a document for a single AWS API operation."""
    command = driver._get_command_table()[service_name]
    if isinstance(command, BasicCommand):
        command_table = command.subcommand_table
    elif isinstance(command, ServiceCommand):
        command_table = command._get_command_table()
    else:
        logger.warning(f'Unknown command type {service_name} {command}')
        return None

    operation = command_table[operation_name]

    help_command = operation.create_help_command()
    event_handler = help_command.EventHandlerClass(help_command)

    # Get description
    event_handler.doc_description(help_command)
    description = _clean_description(help_command.doc.getvalue().decode('utf-8')).strip()

    # Get parameters
    params = {}
    seen_arg_groups = set()
    for arg_name, arg in help_command.arg_table.items():
        if getattr(arg, '_UNDOCUMENTED', False) or arg_name in IGNORED_ARGUMENTS:
            continue
        if arg.group_name in seen_arg_groups:
            continue
        help_command.doc = ReSTDocument()
        if hasattr(event_handler, 'doc'):
            event_handler.doc = help_command.doc
        event_handler.doc_option(help_command=help_command, arg_name=arg_name)
        key = arg.group_name if arg.group_name else arg_name
        params[key] = _clean_text(help_command.doc.getvalue().decode('utf-8').strip())
        params[key] = params[key][:500] if len(params[key]) > 500 else params[key]
        if arg.group_name:
            # To avoid adding arguments like --disable-rollback and --no-disable-rollback separately
            # we need to make sure a group name is only processed once
            # event_handler.doc_option takes care of mentioning all arguments in a group
            # so we can safely skip the remaining arguments in the group
            seen_arg_groups.add(arg.group_name)

    return {
        'command': f'aws {service_name} {operation_name}',
        'description': description,
        'parameters': params,
    }
